import sys
import argparse
from laboro.context.workflow import Context
from laboro.config.manager import Manager as CfgMgr
from laboro.workflow import Workflow
from laboro.logger.manager import Manager as LogMgr


def run(context, cfg_mgr):
  try:
    with Workflow(context=context, **cfg_mgr.workflow_config) as wkf:
      wkf.run()
  except Exception:
    pass


def main(workflows):
  cfg_mgr = CfgMgr(main_config="/etc/laboro/laboro.yml")
  log_mgr = LogMgr()
  context = Context(log_mgr=log_mgr, config_mgr=cfg_mgr)
  log_mgr.logger.log_section("LABORO", "Bootstrapping")
  for workflow_cfg in workflows:
    log_mgr.logger.vault.clear()
    cfg_mgr.workflow_config = workflow_cfg
    run(context=context, cfg_mgr=cfg_mgr)


if __name__ == "__main__":
  parser = argparse.ArgumentParser(description="Run Laboro workflow",
                                   prog="laboro")
  parser.add_argument("-w", "--workflow",
                      metavar="workflows",
                      nargs="+",
                      required=True,
                      help="Run the specified workflows sequentially")
  args = parser.parse_args()
  if not args.workflow:
    parser.print_help()
    sys.exit(1)
  main(args.workflow)



import os
import shutil
import logging
from laboro.error import LaboroError
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Workspace:
  """The ``laboro.workspace.Workspace`` object manages the `laboro.workflow.`workflow`` instance workspace.

  The workflow workspace is a directory named after the workflow name within the **Laboro** global ``workspacedir`` directory.

  Each workflow execution has its own sub-directory  named after the workflow session.

  A ``Workspace`` instance is used to store files within the workflow execution.

  Arguments:
    workspacedir: A string defining the directory where all workspaces are stored.
    workflow: A string defining the workflow name.
    session: A unique string defining the workflow execution session.

  Returns:
    ``laboro.workspace.Workspace``: The workspace for the workflow session.

  Raises:
    ``laboro.error.LaboroError``: When any error such as *OSError* while creating the workspace occurs.
  """
  def __init__(self, workspacedir, workflow, session):
    try:
      self.workspace_dir = os.path.join(workspacedir, workflow)
      self.workspace_path = os.path.join(workspacedir, workflow, session)
      logger.info(f"[+] Creating workspace: {self.workspace_path}")
      os.makedirs(self.workspace_path)
    except Exception as err:
      raise LaboroError(f"[{err.__class__.__name__}] {str(err)}") from err

  def data_to_file(self, filename, data, mode="w",
                   encoding=None, offset=0, whence=0):
    filename = os.path.join(self.workspace_path, filename)
    logger.info(f"Saving file: {filename} ({mode} / {encoding} / ({offset}, {whence}))")
    with open(filename, mode=mode, encoding=encoding) as datafile:
      datafile.seek(offset, whence)
      datafile.write(data)

  def data_from_file(self, filename, length=None, mode="r",
                     encoding=None, offset=0, whence=0):
    filename = os.path.join(self.workspace_path, filename)
    logger.info(f"Reading file: {filename} ({mode} / {encoding} / ({offset}, {whence}))")
    with open(filename, mode=mode, encoding=encoding) as datafile:
      datafile.seek(offset, whence)
      return datafile.read(length)

  def stream_from_file(self, filename, mode="r", encoding=None):
    filename = os.path.join(self.workspace_path, filename)
    logger.info(f"Getting bytestream: {filename} ({mode} / {encoding})")
    return open(filename, mode=mode, encoding=encoding)

  def delete(self):
    """Deletes the instance workspace and all its content.
    This will **not** delete the workflow workspace.

    Raises:
      ``laboro.error.LaboroError``: When any error such as *OSError* while creating the workspace occurs.
    """
    logger.info(f"[+] Deleting workspace: {self.workspace_path}")
    try:
      shutil.rmtree(self.workspace_path)
    except Exception as err:
      raise LaboroError(f"[{err.__class__.__name__}] {str(err)}") from err



import copy
from laboro.context.store import Store
from laboro.logic.processor import Processor


class Context:
  """The ``laboro.context.Context`` class is the base class of ``laboro.context.workflow.Context``, ``laboro.context.step.Context``, ``laboro.context.action.Context`` and ``laboro.context.method.Context``.

  It provides a ``laboro.context.store.Store`` object to store object, variables. It also register `secrets` to `logger`'s `vault` instance.

  Its main purpose is:
    - Register modules to the workflow ``ModuleManager`` instance
    - Instantiate object (with instance secrets registration)
    - Store instantiated objects in its ``Store`` instance

  Arguments:
    logger: A ``laboro.logger.manager.Manager`` instance.
    config_mgr: A ``laboro.config.manager.Manager`` instance.

  Returns:
    ``laboro.context.Context``
  """
  @property
  def log(self):
    return self.log_mgr.logger

  @property
  def store(self):
    return self.store_as_dict()

  def __init__(self, log_mgr, config_mgr):
    self.config_mgr = config_mgr
    self.log_mgr = log_mgr
    self.parent = None
    self.workspace = None
    self.module_mgr = None
    self._store = Store()

  def exit(self, kind, value):
    """Exit the ``Context`` instance.

    The exit method is simply a call to the parent context.
    Subclasses of the Context class may redefine this method.

    Arguments:
      kind: The kind of event that triggered the ``Context`` instance exit.
      value: The `value` of the event that triggered the ``Context`` instance exit.
    """
    if self.parent is not None:
      self.parent.exit(kind, value)

  def put(self, prop, value):
    """Store the ``prop`` variable and its value in the ``Context.store`` ``laboro.context.Store`` instance.

    Arguments:
      prop: A string representing the variable name.
      value: Any type, the value to store.

    Raises:
      ``TypeError``: When ``prop`` is not a string.
    """
    if self.parent is not None:
      self.parent.put(prop, value)
    else:
      self._store.put(prop, value)

  def get(self, prop):
    """Return the value of ``prop`` from the store instance. When ``prop`` is not found in the store, return the ``prop`` value from the parent store if any.

    Arguments:
      prop: A string representing the variable name.

    Raises:
      ``TypeError``: When ``prop`` is not a string.
      ``AttributeError``: When ``prop`` does not exist within the Store instance nor within the parent Context store.
    """
    try:
      return self._store.get(prop)
    except AttributeError as err:
      if self.parent is not None:
        return self.parent.get(prop)
      raise err

  def register_class(self, module, cls):
    """Register to the `module manager` the specified class from the specified module.

    Arguments:
      ``module``: The python module name from which the class is.
      ``class``: The class name.
    """
    self.module_mgr.register_class_from_module(cls, module)

  def instantiate(self, name, module, cls, args):
    """Register the class to the ``laboro.module.ModuleMgr`` instance and store the instance into the store for later use.

    Arguments:
      module: A string specifying the *Python* module name from each retrieve the class.
      cls: A string specifying the class name to instantiate.
      args: A `kwargs` that will be used to instantiate the class.

    Returns:
      An instance of the class ``module``.``cls`` instantiated with the ``args`` `kwargs` as init parameters.

    Raises:
      ``laboro.error.LaboroError``: When the specified module or class ar not available and when the specified ``args`` are incompatibles with the class constructor.
    """
    self.log.info(f"[+] Object instantiation: [{name}] {module}.{cls}")
    obj = self.module_mgr.get_class_from_module(cls=cls, module=module)
    instance_args = self._get_instance_args(args=args)
    with obj(context=self, args=instance_args) as instance:
      self._register_instance_secrets(instance, instance_args)
      self.put(name, instance)

  def register_secret(self, secret):
    """Register the given secret to logger's ``laboro.vault.Vault`` instance.
    """
    self.log.vault.add(secret)

  def store_as_dict(self):
    parent_store = {}
    if self.parent is not None:
      parent_store = self.parent.store_as_dict()
    store = self._store.as_dict()
    for key, value in parent_store.items():
      if key not in store:
        store[key] = value
    return store

  def _register_instance_secrets(self, instance, args):
    if instance.specification.get("args") is not None and args is not None:
      class_args = instance.specification.get("args")
      secret_keys = [arg["name"] for arg in class_args if arg["secret"]]
      secrets = [args[key] for key in args.keys() if key in secret_keys]
      list(map(self.register_secret, secrets))

  def _get_instance_args(self, args=None):
    if args is not None:
      interpolated = copy.deepcopy(args)
      for key, value in interpolated.items():
        interpolated[key] = Processor().process_arg(self, str(value))
      return interpolated
    return dict()



import re
import logging
from laboro.logic.processor import Processor


class Method:
  """The ``laboro.workflow.method.Method`` object is a representation of a   ``laboro.workflow.action.Action`` method.

  Arguments:
    context: The ``laboro.context.Context`` instance used by the workflow.
    name: A string, specifying the method name.
    instance: An instance object derived from the ``laboro.module.Module`` class.
    args: A dictionary representation of the arguments that will be passed to the method call.
    output: String. A variable name in which the return value of the method will be stored for a later use within the workflow.
    when: Optional. A string representation of an expression thant can be evaluate as a boolean.
    loop: Optional. A string describing any iterable object.

  Returns:
    ``laboro.workflow.action.Action``
  """

  @property
  def iterable(self):
    """Get the *iterable* object on which loop, if any.
    This property is evaluated for each call from the ``loop`` attribute.
    """
    if self.loop is not None:
      return Processor().process(self.ctx, str(self.loop))
    return [self.loop]

  @property
  def runnable(self):
    """Get the condition in which the step is runnable.
    This property is evaluated for each call from the ``when`` attribute.
    """
    if self.when is not None:
      return Processor().process(self.ctx, str(self.when))
    return True

  def __init__(self, context, instance, name, args=None,
               output=None, when=None, loop=None):
    self.name = name
    self.ctx = context
    self.instance = instance
    self.when = when
    self.loop = loop
    self.args = args
    self.output = output

  def __enter__(self):
    return self

  def __exit__(self, kind, value, traceback):
    pass

  def run(self):
    self.ctx.log.log_section("METHOD", f"{self.instance}.{self.name}")
    for item in self.iterable:
      self.ctx.set_method_item(item)
      if self.runnable:
        instance = self.ctx.get(self.instance)
        method_args = self.ctx.get_method_args(self.args)
        self.ctx.register_method_secrets(instance,
                                         self.name,
                                         method_args)
        result = getattr(instance, self.name)(**method_args)
        if self.output is not None:
          file_ptn = re.compile(r"\$file\$")
          if re.match(file_ptn, self.output) is not None:
            filename = Processor().process_arg(self.ctx, self.output)
            self.ctx.workspace.data_to_file(filename, f"{result}")
          else:
            self.ctx.put(self.output, result)
      else:
        msg = f"Skipping  {self.name}: Condition not met: {self.when}"
        self.ctx.log.warning(msg)



import logging
from types import SimpleNamespace
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Store:
  """The ``laboro.context.store.Store`` object purpose is to store variables from a ``labor.workflow.Workflow`` instance. It is embedded as an attribute of the ``laboro.context.Context`` object.

  Returns:
    ``laboro.context.store.Store``
  """
  def __init__(self):
    self.storage = SimpleNamespace()
    self.step_item = None
    self.action_item = None
    self.method_item = None

  def put(self, prop, value):
    """Store the ``prop`` variable and its value.

    Arguments:
      prop: A string defining the variable name.
      value: Any object that will be stored as the ``prop`` variable value.

    Raises:
      ``TypeError``: When ``prop`` is not a string.
    """
    setattr(self.storage, prop, value)

  def get(self, prop):
    """Retrieve the ``prop`` variable value.

    Arguments:
      prop: A string defining the variable name.

    Returns:
      Any type depending of the ``prop`` value.

    Raises:
      ``AttributeError``: When ``prop`` does not exist within the Store instance.
      ``TypeError``: When ``prop`` is not a string.
    """
    return getattr(self.storage, prop)

  def clear(self):
    """Clear the Store content.
    """
    self.storage = SimpleNamespace()
    self.step_item = None
    self.action_item = None
    self.method_item = None

  def as_dict(self):
    return {key: value for key, value in self.storage.__dict__.items()}



from laboro.context import Context as BaseContext


class Context(BaseContext):
  """The ``laboro.context.action.Context`` object manages all low level tasks for ``laboro.workflow.Action`` instances. It is instantiated once per ``Action``.

  Its main purpose is:
    - Register modules to the workflow ``ModuleManager`` instance
    - Instantiate object (with instance secrets registration)
    - Set and return `$action_item$` values.

  Arguments:
    ``log_mgr``: A ``laboro.logger.manager.Manager`` instance.
    ``config_mg``r: A ``laboro.config.manager.Manager`` instance.

  Returns:
    ``laboro.context.action.Context``
  """
  def __init__(self, parent):
    super().__init__(log_mgr=parent.log_mgr, config_mgr=parent.config_mgr)
    self.parent = parent
    self.module_mgr = self.parent.module_mgr
    self.workspace = self.parent.workspace

  def set_action_item(self, item):
    self._store.action_item = item

  def get_step_item(self):
    return self.parent.get_step_item()

  def get_action_item(self):
    return self._store.action_item

  def get_method_item(self):
    return None

  def store_as_dict(self):
    store = super().store_as_dict()
    store["step_item"] = self.get_step_item()
    store["action_item"] = self.get_action_item()
    store["method_item"] = None
    return store



import re
import logging
from laboro.logic.parser import Parser
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Processor:
  @staticmethod
  def process(context, statement):
    file_ptn = re.compile(r"\$file\$")
    filepath = f"{context.workspace.workspace_path}/"
    statement = re.sub(file_ptn, filepath, statement)
    datafile_ptn = re.compile(r"\$datafile\$")
    datafile_path = f"{context.workspace.workspace_dir}/"
    statement = re.sub(datafile_ptn, datafile_path, statement)
    return Parser().parse(statement, context.store_as_dict())

  @staticmethod
  def process_arg(context, value):
    file_ptn = re.compile(r"\$file\$")
    filepath = f"{context.workspace.workspace_path}/"
    value = re.sub(file_ptn, filepath, value)
    datafile_ptn = re.compile(r"\$datafile\$")
    datafile_path = f"{context.workspace.workspace_dir}/"
    value = re.sub(datafile_ptn, datafile_path, value)
    return Parser().parse(value, context.store_as_dict())



import logging
from laboro.vault import Vault


class LaboroLogger(logging.Logger):
  """The ``laboro.logger.LaboroLogger`` class overrides the ``logging.Logger`` class.

  It adds a ``laboro.vault.Vault`` instance to the logger object allowing secret management.

  Any secret stored in the vault instance will be redacted before logging.
  """
  def __init__(self, name, level=logging.NOTSET):
    super().__init__(name, level)
    self.vault = Vault()
    self.name = name

  def log_line(self, level=logging.INFO):
    self.log(level=level, msg="-" * 80)

  def log_section(self, prefix, message, level=logging.INFO):
    self.log_line(level=level)
    self.log(level=level, msg=f"[{prefix}] {message}")
    self.log_line(level=level)

  def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
    """Override logging.Logger.makeRecord.
    """
    if self.vault is not None:
      msg = self.vault.protect(msg)
    return super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)



import copy
from laboro.error import LaboroError
from laboro.logic.processor import Processor
from laboro.context import Context as BaseContext


class Context(BaseContext):
  """The ``laboro.context.method.Context`` object manages all low level tasks for ``laboro.workflow.Method`` instances. It is instantiated once per ``Method``.

  Its main purpose are:
    - Register method secrets
    - Set and return `$method_item$` values.

  Arguments:
    ``log_mgr``: A ``laboro.logger.manager.Manager`` instance.
    ``config_mgr``: A ``laboro.config.manager.Manager`` instance.

  Returns:
    ``laboro.context.method.Context``
  """
  def __init__(self, parent):
    super().__init__(log_mgr=parent.log_mgr, config_mgr=parent.config_mgr)
    self.parent = parent
    self.module_mgr = self.parent.module_mgr
    self.workspace = self.parent.workspace

  def register_method_secrets(self, instance, method, args):
    """Register the secrets from the specified arguments ``args``.
    Retrieve the name of the ``instance``'s ``method`` declared as *secret* and register their value in the ``laboro.Vault.vault`` embedded in the ``laboro.logger.LaboroLogger``.

    Arguments:
      instance: An instance of an object derived from the ``laboro.module.Module``.
      method: The ``instance`` method name from which retrieve the secret arguments list.
      args: A kwargs from which retrieve the secret value.

    Raises:
      ``laboro.error.LaboroError`` when the specified ``method`` does not exists.
    """
    if args is not None:
      try:
        method_spec = [meth for meth in instance.specification["methods"] if meth["name"] == method][0]
        if method_spec.get("args") is not None:
          spec_args = method_spec.get("args")
          secret_keys = [arg["name"] for arg in spec_args if arg["secret"]]
          secrets = [args[key]
                     for key in args.keys() if key in secret_keys]
          list(map(self.register_secret, secrets))
      except IndexError as err:
        raise LaboroError(f"UnknownMethodError: Unknown method {instance.__class__.__name__}.{method}") from err

  def get_method_args(self, args):
    if args is not None:
      interpolated = copy.deepcopy(args)
      for key, value in interpolated.items():
        interpolated[key] = Processor().process_arg(self, str(value))
      return interpolated
    return dict()

  def set_method_item(self, item):
    self._store.method_item = item

  def get_step_item(self):
    return self.parent.get_step_item()

  def get_action_item(self):
    return self.parent.get_action_item()

  def get_method_item(self):
    return self._store.method_item

  def store_as_dict(self):
    store = super().store_as_dict()
    store["step_item"] = self.get_step_item()
    store["action_item"] = self.get_action_item()
    store["method_item"] = self.get_method_item()
    return store



import logging
from ast import literal_eval
from asteval import Interpreter
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Parser:
  @staticmethod
  def parse(obj, store):
    res = obj
    if isinstance(obj, str):
      return Parser().eval(obj, store)
    elif isinstance(obj, dict):
      res = {key: Parser().parse(value, store) for key, value in res.items()}
    elif isinstance(obj, (list, tuple)):
      res = [Parser().parse(item, store) for item in res]
      if isinstance(obj, tuple):
        res = tuple(res)
    return res

  @staticmethod
  def eval(expression, store):
    if isinstance(expression, str):
      try:
        res = Interpreter(usersyms=store).eval(expression,
                                               raise_errors=True,
                                               show_errors=False)
      except (KeyError, TypeError, ValueError, SyntaxError, NameError):
        return expression
      if isinstance(res, str):
        return Parser().literal_eval(res)
      return Parser().parse(res, store)
    return expression

  @staticmethod
  def literal_eval(expression):
    try:
      return literal_eval(expression)
    except (KeyError, TypeError, ValueError, SyntaxError, NameError):
      return expression



import os
from laboro.error import LaboroError
from laboro.validator import Validator


class Module:
  """The ``laboro.module.Module`` class is the base class from which each **Laboro** module must be derived from.

  It provides the ``@laboro_method`` decorator which must decorate any **public method** of a derived module.

  ..  code-block:: python

        from laboro.module import Module

        class MyModule(Module):
          def __init__(self, context, args):
            super().__init__(filepath=__file__, context=context, args=args)

        @Module.laboro_method
        def my_public_method(self, arg1, arg2):
          self._private_method(arg1)

        def _my_private_method(self, arg1):
          <do thing>

  Arguments:
    filepath: The file path of the module. From the derived class scope, it should be ``__file__``. The ``filepath`` argument is used to retrieve the module *YAML* specification file.
    context: The ``laboro.context.Context`` associated with the ``Workflow``, ``Step`` or ``Action`` instance instantiating the module. It will be automatically provided.
    args: A kwargs representing the module instance arguments.

  Returns:
    ``laboro.module.Module``

  Raises:
    laboro.error.LaboroError: When the ``args`` kwargs does not validate against the module specification or when the module specification does not validate the meta module specification.
  """
  def __init__(self, filepath, context, args=None):
    self.filepath = filepath
    self.context = context
    self.args = args
    self.specification = None

  @staticmethod
  def laboro_method(func):
    """The ``@laboro.module.Module.laboro_method`` decorator provides the way to validate the module methods arguments against the module method specification. All module derived from the `laboro.module.Module`` **must** decorate all their **public methods**.
    """
    def wrapper(self, *args, **kwargs):
      try:
        method = func.__name__
        methods = self.specification.get("methods") or dict()
        method = [meth for meth in methods if meth["name"] == method][0]
        Validator().validate_method_args(method, kwargs)
        method_args = Validator().set_default_method_args(method, kwargs)
      except IndexError as err:
        raise LaboroError(f"UnknownMethodError: Unknown method: {method}") from err
      return func(self, **method_args)
    return wrapper

  def __enter__(self):
    self.specification = self._validate_spec()
    self._validate_args()
    self._default_args()
    return self

  def __exit__(self, kind, value, traceback):
    pass

  def _validate_spec(self):
    """Load module specification from its YAML data file."""
    base_spec = os.path.join(os.path.dirname(__file__),
                             "schema",
                             "instance.yml")
    spec_file = os.path.join(os.path.dirname(self.filepath),
                             "schema",
                             "specification.yml")
    return Validator().validate_from_files(schema=base_spec,
                                           instance=spec_file)

  def _validate_args(self):
    """Validate arguments against a specification.
    """
    spec_file = os.path.join(os.path.dirname(self.filepath),
                             "schema",
                             "specification.yml")
    if self.args is not None:
      Validator().validate_obj_args(specification=spec_file,
                                    args=self.args)

  def _default_args(self):
    spec_file = os.path.join(os.path.dirname(self.filepath),
                             "schema",
                             "specification.yml")
    self.args = Validator().set_default_obj_args(spec_file, self.args)

  def get_arg_value(self, arg):
    """Get the value of the specified argument.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      The value of the searched argument. The type of the return value depend of the value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if self.args is not None and arg in self.args:
      return self.args[arg]
    msg = {"type": "UnknownModuleArgError",
           "message": f"Module argument not found: {arg}"}
    raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def get_arg_value_as_string(self, arg):
    """Get the value of the specified argument as a string.

    Arguments:
      arg: The argument to get the value from.

    Returns:
      ``str``: The string representation of value of the searched argument.

    Raises:
      ``laboro.error.LaboroError``: When the specified argument is not known to the module.
    """
    if self.args is not None and arg in self.args:
      return str(self.args[arg])
    msg = {"type": "UnknownModuleArgError",
           "message": f"Module argument not found: {arg}"}
    raise LaboroError(f"[{msg['type']}] {msg['message']}")



import uuid
from laboro.context.step import Context as StepContext
from laboro.workflow.step import Step


class Workflow:
  """The ``laboro.workflow.Workflow`` object is the main class for the workflow representation.
  It load all configuration needed, set alk objects such History, Vault and workspace and run according to its configuration.

  The Workflow object provides a runtime context that will handle log, history, vault, and workspace, etc.

  Arguments:
    name: A string representing the workflow name.
    context: A ``laboro.context.Context`` instance.
    name: A string specifying the name of the workflow
    packages: A list of package names.
    workspace: A dictionary representation of a workspace configuration (see ``laboro.workspace.Workspace`` for further dertails).
    steps: A list of dictionary representation of ``laboro.workflow.Step``

  Returns:
    ``laboro.workflow.Workflow``: A Workflow object.

  ..  code-block:: python

      from laboro.vault import Vault()
      from laboro.log.manager import Manager as LogMgr
      from laboro.context import Context
      from laboro.workflow import Workflow

      cfg_mgr = CfgMgr(main_config="/etc/laboro/laboro.yml")
      context = Context(logger=LogMgr(Vault()), config_mgr=cfg_mgr)
      self.ctx.log.log_section("LABORO", "Bootstrapping")
      self.ctx.log.vault.clear()
      cfg_mgr.workflow_config = "my_workflow.yml"
      workflow_config = cfg_mgr.workflow_config
      with Workflow(context=context, **workflow_config) as wkf:
        wkf.run()
        ...
  """
  def __init__(self, context, name, packages, workspace, steps, instances=None):
    self.name = name.replace(" ", "_")
    self.packages = packages
    self.workspace = workspace
    self.steps = steps
    self.instances = instances
    if instances is None:
      self.instances = list()
    self.ctx = context.reset(session=str(uuid.uuid4()),
                             workspace_cfg=self.workspace)

  def __enter__(self):
    return self

  def __exit__(self, kind, value, traceback):
    self.ctx.exit(kind, value)

  def _instantiate(self, instance):
    self.ctx.register_class(module=instance.get("module"),
                            cls=instance.get("class"))
    self.ctx.instantiate(module=instance.get("module"),
                         cls=instance.get("class"),
                         name=instance.get("name"),
                         args=instance.get("args"))

  def run(self):
    """Run the workflow."""
    for instance in self.instances:
      self._instantiate(instance)
    for wkf_step in self.steps:
      with Step(StepContext(parent=self.ctx),
                **wkf_step) as step:
        step.run()
    self.ctx.log.log_line()



import os
from laboro.context import Context as Ctx
from laboro.history import History
from laboro.workspace import Workspace
from laboro.module.manager import Manager as ModuleMgr
from laboro.error.handler import Handler as ErrorHandler


class Context(Ctx):
  """The ``laboro.context.workflow.Context`` object manages all low level tasks for ``laboro.workflow.Workflow`` instances. It is instantiated once by the main **Laboro** process and is passed from **Workflow** instances to **Workflow** instances.

  It loads the main **Laboro** configuration file and initiate all needed **Laboro** submodules.

  Arguments:
    ``log_mgr``: A ``laboro.logger.manager.Manager`` instance.
    ``config_mgr``: A ``laboro.config.manager.Manager`` instance.

  Raises:
    ``laboro.error.LaboroError``: If the main **Laboro** configuration file is not found or invalid.

  Returns:
    ``laboro.context.workflow.Context``
  """
  def __init__(self, log_mgr, config_mgr):
    super().__init__(log_mgr, config_mgr)
    self.delete_workspace_on_exit = True
    self.history = None
    self.module_mgr = ModuleMgr()
    self.config_mgr = config_mgr
    self.workspacedir = self.config_mgr.get_parameter("main",
                                                      "$.laboro.workspacedir")
    self.histdir = self.config_mgr.get_parameter("main",
                                                 "$.laboro.histdir")
    self.workflowdir = self.config_mgr.get_parameter("main",
                                                     "$.laboro.workflowdir")
    self.logdir = self.config_mgr.get_parameter("main",
                                                "$.laboro.log.dir")
    self.loglevel = self.config_mgr.get_parameter("main",
                                                  "$.laboro.log.level")
    self.workflow_name = None
    self.workflow_session = None
    self.workspace = None
    self.parent = None

  def reset(self, session, workspace_cfg):
    """Reset the ``laboro.context.Context`` instance with the new workflow configuration and session.

    Arguments:
      session: A unique string defining a workflow session. Usually a string representation of a uuid.uuid4() instance.
      workspace_cfg: A dict representing the workspace configuration. See ``laboro.workspace.Workspace`` and ``laboro.workflow.Workflow`` for workspace configuration details.

    Raises:
      ``laboro.error.LaboroError``: For various reasons linked to the submodules instantiations.

    Returns:
      ``laboro.context.Context``: The ``Context`` instance reset with the new parameters.
    """
    if self.parent is None:
      try:
        self.workflow_session = session
        self.workflow_name = self.config_mgr.get_parameter("workflow", "$.name")
        self.workflow_name = self.workflow_name.replace(" ", "_")
        self._configure_logger()
        start_msg = f"Started {self.workflow_name} / {self.workflow_session}"
        self.log.log_section("WORKFLOW", start_msg)
        self.log.vault.clear()
        self._store.clear()
        self.delete_workspace_on_exit = workspace_cfg["delete_on_exit"]
        self.workspace = Workspace(workspacedir=self.workspacedir,
                                   workflow=self.workflow_name,
                                   session=self.workflow_session)
        history_path = os.path.join(self.histdir, f"{self.workflow_name}.db")
        self.history = History(filename=history_path,
                               workflow=self.workflow_name,
                               session=self.workflow_session,
                               params=self.config_mgr.workflow_config)
        self.history.enter()
        self.install_packages()
        return self
      except Exception as err:
        ErrorHandler().handle_error(self.workflow_name,
                                    self.workflow_session,
                                    err.__class__,
                                    str(err))

  def exit(self, kind, value):
    """Exit the ``Context`` instance.
    When the instance exit, it triggers the ``laboro.workspace.Workspace``, ``laboro.history.History`` instances exit and close the log linked to the ``laboro.workflow.Workflow`` instance associated with the ``Context`` instance. It also register the exit/error code.

    Arguments:
      kind: The kind of event that triggered the ``Context`` instance exit.
      value: The `value` of the event that triggered the ``Context`` instance exit.
    """
    if self.delete_workspace_on_exit:
      self.workspace.delete()
    self.history.exit(kind, value)
    ErrorHandler().handle_error(self.workflow_name,
                                self.workflow_session,
                                kind,
                                value)
    self.log_mgr.remove_file_handler(self.workflow_session)

  def _configure_logger(self):
    self.log_mgr.add_file_handler(self.logdir,
                                  self.workflow_name,
                                  self.workflow_session)
    self.log_mgr.set_log_level(self.loglevel)

  def install_packages(self):
    """Install all packages listed in the workflow configuration file.
    """
    packages = self.config_mgr.get_parameter("workflow", "$.packages")
    for pkg in packages:
      self.module_mgr.install_package(pkg)



import copy
import yaml
import logging
import builtins
from jsonschema import validate
from jsonschema.exceptions import ValidationError, SchemaError
from laboro.error import LaboroError
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Validator:
  """The ``laboro.validator.Validator`` object purpose is to validate `json` object against `jsonschema` specification given as `YAML` files.

  This object is instantiated by some **Laboro** core components and is **not intended to be used directly**.
  """

  def validate_from_files(self, schema, instance):
    """Validate the given instance against the given schema.

    Arguments:
    ``schema``: A file path to a *YAML* representation of a `jsonschema` representing the schema against which the ``instance`` should be validated.

    ``instance``: A file path to a *YAML* representation of a `json` representing the instance to validate.

    Raises:
      ``laboro.error.Error``: Whenever the schema or the instance file can't be loaded or are not valid *YAML* files. Error is also raised when the given ``schema`` is not a valid `jsonschema` or when ``instance`` can not validate against the given ``schema``.
    """
    try:
      schema_yml = self._load_file(schema)
      instance_yml = self._load_file(instance)
      validate(instance=instance_yml, schema=schema_yml)
      return instance_yml
    except ValidationError as err:
      msg = {"type": "InvalidSchemaInstanceError",
             "message": f"Instance is invalid: {err.message}"}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err
    except SchemaError as err:
      msg = {"type": "InvalidSchemaError",
             "message": f"Schema is invalid: {err.message}"}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def validate_method_args(self, method, args):
    """Validate the method arguments against the method specification.

    Arguments:
      method: A dict describing the method specification from the module.
      args: A dict representing the method argument and their values.

    Raises:
      ``laboro.error.LaboroError``: When the given ``args`` does not validate against the specification.
    """
    try:
      if method.get("args") is None:
        if len(args) > 0:
          msg = {"type": "TooManyArgsError",
                 "message": f"Method does not accept args: {method['name']}"}
          logger.critical(f"[{msg['type']}] {msg['message']}")
          raise LaboroError(f"[{msg['type']}] {msg['message']}")
        return
      if not isinstance(args, dict):
        msg = f"[MethodArgError]: Expected dict, received {type(args)}"
        logger.critical(msg)
        raise LaboroError(msg)
      if not self._validate_required(args, method):
        msg = "[MissingRequiredArgError] Missing required argument."
        logger.critical(msg)
        raise LaboroError(msg)
      for arg, value in args.items():
        if arg in [arg["name"] for arg in method["args"]]:
          self._validate_type(arg, value, method)
          self._validate_implied(arg, args, method)
          self._validate_excluded(arg, args, method)
        else:
          msg = {"type": "UnknownMethodArgError",
                 "message": f"Unknown argument: {arg}"}
          logger.critical(f"[{msg['type']}] {msg['message']}")
          raise LaboroError(f"[{msg['type']}] {msg['message']}")
    except AttributeError as err:
      msg = {"type": "InvalidMethodSpecTypeError",
             "message": f"Specification must be dict. Received: {type(method)}"}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def validate_obj_args(self, specification, args):
    """Validate the instantiation arguments against the object specification.

    Arguments:
      specification: A file path to a *YAML* representation of a `jsonschema` representing the schema against which the ``args`` should be validated.

      args: A dict representing the object instantiation argument and their values.

    Raises:
      ``laboro.error.LaboroError``: When the given ``args`` does not validate against the specification.
    """
    if not isinstance(args, dict):
      msg = f"[ModuleArgError]: Expected a dict as args, received {type(args)}"
      logger.critical(msg)
      raise LaboroError(msg)
    obj = self._load_file(specification)
    if obj.get("args") is None:
      if len(args) > 0:
        msg = {"type": "TooManyArgsError",
               "message": f"Object constructor does not accept args: {obj['class']}"}
        logger.critical(f"[{msg['type']}] {msg['message']}")
        raise LaboroError(f"[{msg['type']}] {msg['message']}")
      return
    if not self._validate_required(args, obj):
      msg = "[MissingRequiredArgError] Missing required argument."
      logger.critical(msg)
      raise LaboroError(msg)
    for arg, value in args.items():
      if arg in [arg["name"] for arg in obj["args"]]:
        self._validate_type(arg, value, obj)
        self._validate_implied(arg, args, obj)
        self._validate_excluded(arg, args, obj)
      else:
        msg = {"type": "UnknownObjArgError",
               "message": f"Unknown argument: {arg}"}
        logger.critical(f"[{msg['type']}] {msg['message']}")
        raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def set_default_obj_args(self, specification, args):
    obj = self._load_file(specification)
    return self._set_default_args(obj, args)

  def set_default_method_args(self, method, args):
    return self._set_default_args(method, args)

  def _set_default_args(self, obj, args):
    if obj.get("args") is not None:
      default_args = copy.deepcopy(obj.get("args"))
      for arg in default_args:
        if arg.get("default") is not None:
          arg["value"] = arg["default"]
        if args.get(arg["name"]) is not None:
          arg["value"] = args.get(arg["name"])
      return {arg["name"]: arg["value"] for arg in default_args if arg.get("value") is not None}
    return args

  def _load_file(self, filepath):
    try:
      with open(filepath, mode="r", encoding="utf-8") as schema:
        return yaml.load(schema, Loader=yaml.FullLoader)
    except FileNotFoundError as err:
      msg = {"type": "FileNotFoundError",
             "message": f"File not found: {filepath}"}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err
    except(yaml.scanner.ScannerError,
           yaml.constructor.ConstructorError) as err:
      msg = {"type": "InvalidYamlError",
             "message": f"YAML file is invalid: {filepath}"}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def _validate_required(self, arguments, specification):
    try:
      required = sorted([arg["name"] for arg in list(filter(lambda arg: arg["required"], specification["args"]))])
      return sorted([arg for arg in arguments if arg in required]) == required
    except Exception as err:
      raise LaboroError(f"[InvalidSpecError] Unable to validate required args: {err}") from err

  def _validate_type(self, arg, value, specification):
    arg_type = [aarg["type"] for aarg in specification["args"] if aarg["name"] == arg][0]
    required = [aarg["required"] for aarg in specification["args"] if aarg["name"] == arg][0]
    if required or value is not None:
      try:
        allowed_types = arg_type if isinstance(arg_type, list) else [arg_type]
        allow = [isinstance(value, getattr(builtins, allowed))
                 for allowed in allowed_types]
        if not any(allow):
          msg = {"type": "BadArgTypeError",
                 "message": f"Expected type one of {allowed_types} / received {type(value)} for {arg}."}
          logger.critical(f"[{msg['type']}] {msg['message']}")
          raise LaboroError(f"[{msg['type']}] {msg['message']}")
      except AttributeError as err:
        msg = {"type": "UnknownTypeError",
               "message": f"Unknown type '{arg_type}' for argument {arg}."}
        logger.critical(f"[{msg['type']}] {msg['message']}")
        raise LaboroError(f"[{msg['type']}] {msg['message']}") from err

  def _validate_implied(self, arg, arguments, specification):
    try:
      implied = sorted([aarg["implied"] for aarg in list(filter(lambda arg: arg["implied"], specification["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      implied = list()
    if sorted([iarg for iarg in arguments if iarg in implied]) != implied:
      msg = {"type": "MissingImpliedArgError",
             "message": f"All args of {implied} are required by {arg}."}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}")

  def _validate_excluded(self, arg, arguments, specification):
    try:
      excluded = sorted([aarg["excluded"] for aarg in list(filter(lambda arg: arg["excluded"], specification["args"])) if aarg["name"] == arg])[0]
    except IndexError:
      excluded = list()
    if len([earg for earg in arguments if earg in excluded]) > 0:
      msg = {"type": "ExcludedArgError",
             "message": f"None of {' or '.join(excluded)} can be used alongside {arg}."}
      logger.critical(f"[{msg['type']}] {msg['message']}")
      raise LaboroError(f"[{msg['type']}] {msg['message']}")



import os
import sys
import logging
import coloredlogs
from laboro.logger import LaboroLogger


class Manager:
  """The purpose of the ``laboro.logger.manager.Manager`` object is to configure the logger used in the whole **Laboro** package and all its subpackages.
  """

  terms = ["linux", "xterm", "xterm-256color"]
  console_fmt = '[%(asctime)s] [%(levelname)-8s] %(message)s'
  file_fmt = '[%(asctime)s] [%(levelname)-8s] %(message)s'
  datefmt = '%Y%m%d %H:%M:%S'
  c_field_styles = {'asctime': {'color': 'magenta'},
                    'hostname': {'color': 'magenta'},
                    'levelname': {'bold': True, 'color': 'white'},
                    'programname': {'color': 'cyan'},
                    'name': {'color': 'cyan'}}
  c_level_styles = {'debug': {'color': 'white'},
                    'info': {'color': 'green'},
                    'warning': {'color': 'yellow'},
                    'error': {'color': 'red'},
                    'critical': {'bold': True, 'color': 'red'}}
  f_field_styles = {'asctime': {'color': 'magenta'},
                    'hostname': {'color': 'magenta'},
                    'levelname': {'color': 'white'},
                    'programname': {'color': 'cyan'},
                    'name': {'color': 'cyan'}}
  f_level_styles = {'debug': {'color': 'white'},
                    'info': {'color': 'green'},
                    'warning': {'color': 'yellow'},
                    'error': {'color': 'red'},
                    'critical': {'color': 'red'}}

  def __init__(self):
    logging.setLoggerClass(LaboroLogger)
    self.handlers = {}
    self._set_default()
    self.logger = logging.getLogger("laboro.main")

  def _set_default(self):
    """Configure the logger default configuration to *stdout*.

    If the ``TERM`` environment variable is defined and is one of "linux", "xterm" or "xterm-256color" the default logger will use ``coloredlogs`` to enhance log readability with colors.

    The default logging level is set to ``INFO``.
    """
    console_handler = logging.StreamHandler(sys.stdout)
    if "TERM" in os.environ and os.environ["TERM"] in self.terms:
      console_handler.setFormatter(
          coloredlogs.ColoredFormatter(fmt=self.console_fmt,
                                       datefmt=self.datefmt,
                                       level_styles=self.c_level_styles,
                                       field_styles=self.c_field_styles))
    else:
      console_handler.setFormatter(logging.Formatter(fmt=self.console_fmt,
                                                     datefmt=self.datefmt))
    logging.basicConfig(level=logging.INFO,
                        handlers=[console_handler])

  def set_log_level(self, level):
    """Set the log level

    Arguments:
      level: A string specifying the desired log level. It **must** be one of the ``logging`` *Python* module standard levels (``DEBUG``, ``INFO``, ``WARNING``, ``ERROR``, ``CRITICAL``).

    Raises:
      ``ValueError``: When the given ``level`` is not a valid level name.
    """
    self.logger.setLevel(level)

  def add_file_handler(self, logdir, workflow, session):
    """Adds a file handler to the logger.

    Arguments:
      logdir: A string defining the log directory in which all logs will be stored.
      workflow: A string specifying workflow's name.
      session: A unique string defining the workflow execution session.
    """
    log_dname = os.path.join(logdir, workflow)
    os.makedirs(log_dname, exist_ok=True)
    log_fname = os.path.join(log_dname, f"{workflow}_{session}.log")
    self.handlers[session] = logging.FileHandler(log_fname)
    self.handlers[session].setFormatter(
        coloredlogs.ColoredFormatter(fmt=self.file_fmt,
                                     datefmt=self.datefmt,
                                     level_styles=self.f_level_styles,
                                     field_styles=self.f_field_styles))
    self.logger.addHandler(self.handlers[session])

  def remove_file_handler(self, session):
    """Removes the file handler associated with the specified session.

    Arguments:
      session: A unique string specifying a workflow session.
    """
    self.logger.removeHandler(self.handlers[session])

  @staticmethod
  def print_logfile(logdir, workflow, session):
    """Print to *stdout* the log file content for the specified workflow session.

    Arguments:
      logdir: A string defining the log directory in which all logs will be stored.
      workflow: A string specifying the workflow name.
      session: A unique string defining the workflow execution session to retrieve the log for
    """
    log_dname = os.path.join(logdir, workflow)
    log_fname = os.path.join(log_dname, f"{workflow}_{session}.log")
    with open(log_fname, "r", encoding="utf-8") as log:
      print(log.read())



from laboro.context.action import Context as ActionContext
from laboro.logic.processor import Processor
from laboro.workflow.action import Action


class Step:
  """The ``laboro.workflow.step.Step`` object is the object representation of a workflow step.

  Arguments:
    context: The ``laboro.context.Context`` instance used by the workflow.
    name: A string, specifying the step name.
    actions: A list, the list of the step actions
    when: Optional. A string representation of an expression thant can be evaluate as a boolean.
    loop: Optional. A string describing any iterable object.

  Returns:
    ``laboro.workflow.step.Step``

  """
  @property
  def iterable(self):
    """Get the *iterable* object on which loop, if any.
    This property is evaluated for each call from the ``loop`` attribute.
    """
    if self.loop is not None:
      return Processor().process(self.ctx, str(self.loop))
    return [self.loop]

  @property
  def runnable(self):
    """Get the condition in which the step is runnable.
    This property is evaluated for each call from the ``when`` attribute.
    """
    if self.when is not None:
      return Processor().process(self.ctx, str(self.when))
    return True

  def __init__(self, context, name, actions,
               instances=None, when=None, loop=None):
    self.name = name
    self.ctx = context
    self.instances = instances
    if instances is None:
      self.instances = list()
    self.when = when
    self.loop = loop
    self.actions = actions

  def __enter__(self):
    return self

  def __exit__(self, kind, value, traceback):
    pass

  def _instantiate(self, instance):
    self.ctx.register_class(module=instance.get("module"),
                            cls=instance.get("class"))
    self.ctx.instantiate(module=instance.get("module"),
                         cls=instance.get("class"),
                         name=instance.get("name"),
                         args=instance.get("args"))

  def run(self):
    """Run the step.
    Calling ``laboro.workflow.step.Step.run()`` will run all actions within the step if the ``runnable`` property evaluate to ``True``.
    The actions within the step will be executed for each item in the ``iterable`` property.
    """
    self.ctx.log.log_section("STEP", self.name)
    for instance in self.instances:
      self._instantiate(instance)
    for item in self.iterable:
      self.ctx.set_step_item(item)
      if self.runnable:
        for step_action in self.actions:
          with Action(ActionContext(parent=self.ctx),
                      **step_action) as action:
            action.run()
      else:
        msg = f"Skipping {self.name}: Condition not met: {self.when}"
        self.ctx.log.warning(msg)



import re


class Vault:
  """The ``laboro.vault.Vault`` class manages sensible informations that should not appear in log files.

  **Note:** Even if they will not appear in log files or `stdout`, **data stored in Vault are NOT encrypted** in the memory and are still recoverable by object inspection and debugger tools."""

  @property
  def protected(self):
    return "******"

  def __init__(self):
    self.secrets = list()
    self.pattern = None

  def add(self, string):
    """Adds a secret to the vault.

    Arguments:
      string: A string that should never appears in the logs.
    """
    self.secrets.append(str(string))
    self.pattern = re.compile(f"{'|'.join(self.secrets)}")

  def is_secret(self, string):
    """Checks if a string is registered is the Vault.

    Arguments:
      string: The string to check.

    Returns:
      ``bool``: *True* if ``string`` is a registered secret in this Vault, *False* otherwise.
    """
    return string in self.secrets

  def protect(self, string):
    """Redacts any string containing one of the secrets registered in this Vault.

    Arguments:
      string: The string to redact if necessary.

    Returns:
      ``str``: The ``string`` with all registered secrets redacted.
    """
    if self.pattern is not None:
      return re.sub(self.pattern, self.protected, str(string))
    return string

  def clear(self):
    """Clear the vault. All secret stored in vault will be deleted.
    """
    self.secrets = list()
    self.pattern = None



import re
import sys
import logging
import importlib
import subprocess
from laboro.error import LaboroError
from laboro.logger import LaboroLogger

logging.setLoggerClass(LaboroLogger)
logger = logging.getLogger("laboro.main")


class Manager:
  """The ``laboro.module.manager.Manager`` object manages all aspect of module handling such as package installation, module loading and module registration.
  """
  modules = dict()

  def install_package(self, package):
    """Install the specified *Python* package.

    Arguments:
      package: A string specifying the *Python* package name to install. i.e ``laboro_database``.

    Raises:
      ``laboro.error.LaboroError``: When the specified package is not available.
    """
    try:
      reg = re.compile("^laboro_.*$|^laboro-.*$")
      if re.match(reg, package) is not None:
        logger.info(f"[+] Installing package: {package}")
        subprocess.check_call([sys.executable, "-m", "pip", "-qqq", "install", package])
      else:
        logger.critical(f"Package is not a valid Laboro package: {package}")
        raise LaboroError(f"InvalidPackageNameError: Package is not a valid Laboro package: {package}")
    except subprocess.CalledProcessError as err:
      logger.critical(f"Unable to install Laboro package: {package}")
      raise LaboroError(f"InstallPackageError: Unable to install Laboro package: {package}") from err

  def register_class_from_module(self, cls, module):
    """Registers the ``cls`` class from the python module whose name is specified by the ``module`` argument.

    Once registered the specified ``cls`` will be accessible through the ``modules`` attribute.

    Registering a module is used to import external Laboro modules.

    To prevent abusive standard Python packages imports, module name defined by the ``module`` argument must match the following regular expression: ``^laboro_.*``.

    Arguments:
      cls: A string specifying the class to import from ``module``.
      module: A string specifying the module name from which import the ``cls`` class. The ``module`` name  **must** match ``^laboro_.*``.

    Raises:
      ``laboro.error.LaboroError``: When an ImportError occurred while attempting to register an unknown class or a class from an unknown module.

    In the following example illustrate how to import the ``PgClient`` class from the ``laboro_database.clients`` package.

    ..  code-block:: python

      from laboro.module.manager import Manager

      module_mgr = Manager()
      module_mgr.install_package(laboro_database)
      module_mgr.register(cls="PgClient", module="laboro_database.clients")

    """
    try:
      mod = importlib.import_module(module)
      if not hasattr(mod, cls):
        raise LaboroError(f"ClassNotFoundError: No such class: {module}.{cls}")
      if self.modules.get(module) is not None:
        if self.modules[module].get(cls) is None:
          logger.info(f"[+] Registering class: {module}.{cls}")
          self.modules[module][cls] = getattr(mod, cls)
      else:
        logger.info(f"[+] Registering class: {module}.{cls}")
        self.modules[module] = dict()
        self.modules[module][cls] = getattr(mod, cls)
    except ImportError as err:
      logger.critical(f"No such module: {module}")
      raise LaboroError(f"ModuleNotFoundError: No such module: {module}") from err

  def get_class_from_module(self, cls, module):
    """Return the specified registered class.

    Arguments:
      cls: A string specifying the class to retrieve.
      module: A string specifying the module to retrieve the class from.

    Raises:
      ``laboro.error.LaboroError``: When the specified module or class is unregistered.
    """
    try:
      return self.modules[module][cls]
    except KeyError as err:
      logger.critical(f"No such class: {module}.{cls}")
      raise LaboroError(f"UnregisteredClassError: No such class: {module}.{cls}") from err
